@@ -4,7 +4,10 @@ use std::{
     rc::Rc,
 };
 
-use crate::{sys::TokenFactory, Poll, Readiness, Token};
+use log::trace;
+
+pub use crate::loop_logic::EventIterator;
+use crate::{sys::TokenFactory, Poll, Readiness, RegistrationToken, Token};
 
 pub mod channel;
 #[cfg(feature = "executor")]
@@ -12,7 +15,7 @@ pub mod channel;
 pub mod futures;
 pub mod generic;
 pub mod ping;
-#[cfg(target_os = "linux")]
+#[cfg(all(target_os = "linux", feature = "signals"))]
 #[cfg_attr(docsrs, doc(cfg(target_os = "linux")))]
 pub mod signals;
 pub mod timer;
@@ -31,7 +34,7 @@ pub mod transient;
 /// sources. For example, if every child source returns `Continue`, the result
 /// will be `Continue`, but the parent source might still need to return
 /// `Reregister` or something else depending on any additional logic it uses.
-#[derive(Debug, PartialEq, Copy, Clone)]
+#[derive(Copy, Clone, Debug, PartialEq, Eq)]
 pub enum PostAction {
     /// Continue listening for events on this source as before
     Continue,
@@ -85,10 +88,17 @@ impl BitOrAssign for PostAction {
 ///
 /// In case your event source needs to do some special processing before or after a
 /// polling session occurs (to prepare the underlying source for polling, and cleanup
-/// after that), you can override the `pre_run` and `post_run`, that do nothing by
-/// default. Depending on the underlying events, `process_events` may be invoked once,
-/// several times, or none at all between `pre_run` and `post_run` are called, but when
-/// it is invoked, it'll always be between those two.
+/// after that), you can override [`NEEDS_EXTRA_LIFECYCLE_EVENTS`] to `true`.
+/// For all sources for which that constant is `true`, the methods [`before_sleep`] and
+/// [`before_handle_events`] will be called.
+/// [`before_sleep`] is called before the polling system performs a poll operation.
+/// [`before_handle_events`] is called before any process_events methods have been called.
+/// This means that during `process_events` you can assume that all cleanup has occured on
+/// all sources.
+///
+/// [`NEEDS_EXTRA_LIFECYCLE_EVENTS`]: EventSource::NEEDS_EXTRA_LIFECYCLE_EVENTS
+/// [`before_sleep`]: EventSource::before_sleep
+/// [`before_handle_events`]: EventSource::before_handle_events
 pub trait EventSource {
     /// The type of events generated by your source.
     type Event;
@@ -157,25 +167,38 @@ pub trait EventSource {
     /// [`Poll::unregister`](crate::Poll#method.unregister) method.
     fn unregister(&mut self, poll: &mut Poll) -> crate::Result<()>;
 
-    /// Notification that a polling session is going to start
+    /// Whether this source needs to be sent the [`EventSource::before_sleep`]
+    /// and [`EventSource::before_handle_events`] notifications. These are opt-in because
+    /// they require more expensive checks, and almost all sources will not need these notifications
+    const NEEDS_EXTRA_LIFECYCLE_EVENTS: bool = false;
+    /// Notification that a single `poll` is about to begin
     ///
-    /// You can generate events from this method as you would from `process_events`.
-    fn pre_run<F>(&mut self, _callback: F) -> crate::Result<()>
-    where
-        F: FnMut(Self::Event, &mut Self::Metadata) -> Self::Ret,
-    {
-        Ok(())
-    }
-
-    /// Notification that the current polling session ended
+    /// Use this to perform operations which must be done before polling,
+    /// but which may conflict with other event handlers. For example,
+    /// if polling requires a lock to be taken
     ///
-    /// You can generate events from this method as you would from `process_events`.
-    fn post_run<F>(&mut self, _callback: F) -> crate::Result<()>
-    where
-        F: FnMut(Self::Event, &mut Self::Metadata) -> Self::Ret,
-    {
-        Ok(())
+    /// If this returns Ok(Some), this will be treated as an event arriving in polling, and
+    /// your event handler will be called with the returned `Token` and `Readiness`.
+    /// Polling will however still occur, but with a timeout of 0, so additional events
+    /// from this or other sources may also be handled in the same iterations.
+    /// The returned `Token` must belong to this source
+    // If you need to return multiple synthetic events from this notification, please
+    // open an issue
+    fn before_sleep(&mut self) -> crate::Result<Option<(Readiness, Token)>> {
+        Ok(None)
     }
+    /// Notification that polling is complete, and [`EventSource::process_events`] will
+    /// be called with the given events for this source. The iterator may be empty,
+    /// which indicates that no events were generated for this source
+    ///
+    /// Please note, the iterator excludes any synthetic events returned from
+    /// [`EventSource::before_sleep`]
+    ///
+    /// Use this to perform a cleanup before event handlers with arbitrary
+    /// code may run. This could be used to drop a lock obtained in
+    /// [`EventSource::before_sleep`]
+    #[allow(unused_variables)]
+    fn before_handle_events(&mut self, events: EventIterator<'_>) {}
 }
 
 /// Blanket implementation for boxed event sources. [`EventSource`] is not an
@@ -214,18 +237,14 @@ impl<T: EventSource> EventSource for Box<T> {
         T::unregister(&mut **self, poll)
     }
 
-    fn pre_run<F>(&mut self, callback: F) -> crate::Result<()>
-    where
-        F: FnMut(Self::Event, &mut Self::Metadata) -> Self::Ret,
-    {
-        T::pre_run(&mut **self, callback)
+    const NEEDS_EXTRA_LIFECYCLE_EVENTS: bool = T::NEEDS_EXTRA_LIFECYCLE_EVENTS;
+
+    fn before_sleep(&mut self) -> crate::Result<Option<(Readiness, Token)>> {
+        T::before_sleep(&mut **self)
     }
 
-    fn post_run<F>(&mut self, callback: F) -> crate::Result<()>
-    where
-        F: FnMut(Self::Event, &mut Self::Metadata) -> Self::Ret,
-    {
-        T::post_run(&mut **self, callback)
+    fn before_handle_events(&mut self, events: EventIterator) {
+        T::before_handle_events(&mut **self, events)
     }
 }
 
@@ -266,24 +285,21 @@ impl<T: EventSource> EventSource for &mut T {
         T::unregister(&mut **self, poll)
     }
 
-    fn pre_run<F>(&mut self, callback: F) -> crate::Result<()>
-    where
-        F: FnMut(Self::Event, &mut Self::Metadata) -> Self::Ret,
-    {
-        T::pre_run(&mut **self, callback)
+    const NEEDS_EXTRA_LIFECYCLE_EVENTS: bool = T::NEEDS_EXTRA_LIFECYCLE_EVENTS;
+
+    fn before_sleep(&mut self) -> crate::Result<Option<(Readiness, Token)>> {
+        T::before_sleep(&mut **self)
     }
 
-    fn post_run<F>(&mut self, callback: F) -> crate::Result<()>
-    where
-        F: FnMut(Self::Event, &mut Self::Metadata) -> Self::Ret,
-    {
-        T::post_run(&mut **self, callback)
+    fn before_handle_events(&mut self, events: EventIterator) {
+        T::before_handle_events(&mut **self, events)
     }
 }
 
 pub(crate) struct DispatcherInner<S, F> {
     source: S,
     callback: F,
+    needs_additional_lifecycle_events: bool,
 }
 
 impl<Data, S, F> EventDispatcher<Data> for RefCell<DispatcherInner<S, F>>
@@ -301,50 +317,75 @@ where
         let DispatcherInner {
             ref mut source,
             ref mut callback,
+            ..
         } = *disp;
+        trace!(
+            "[calloop] Processing events for source type {}",
+            std::any::type_name::<S>()
+        );
         source
             .process_events(readiness, token, |event, meta| callback(event, meta, data))
             .map_err(|e| crate::Error::OtherError(e.into()))
     }
 
-    fn register(&self, poll: &mut Poll, token_factory: &mut TokenFactory) -> crate::Result<()> {
-        self.borrow_mut().source.register(poll, token_factory)
+    fn register(
+        &self,
+        poll: &mut Poll,
+        additional_lifecycle_register: &mut AdditionalLifecycleEventsSet,
+        token_factory: &mut TokenFactory,
+    ) -> crate::Result<()> {
+        let mut this = self.borrow_mut();
+
+        if this.needs_additional_lifecycle_events {
+            additional_lifecycle_register.register(token_factory.registration_token());
+        }
+        this.source.register(poll, token_factory)
     }
 
-    fn reregister(&self, poll: &mut Poll, token_factory: &mut TokenFactory) -> crate::Result<bool> {
+    fn reregister(
+        &self,
+        poll: &mut Poll,
+        additional_lifecycle_register: &mut AdditionalLifecycleEventsSet,
+        token_factory: &mut TokenFactory,
+    ) -> crate::Result<bool> {
         if let Ok(mut me) = self.try_borrow_mut() {
             me.source.reregister(poll, token_factory)?;
+            if me.needs_additional_lifecycle_events {
+                additional_lifecycle_register.register(token_factory.registration_token());
+            }
             Ok(true)
         } else {
             Ok(false)
         }
     }
 
-    fn unregister(&self, poll: &mut Poll) -> crate::Result<bool> {
+    fn unregister(
+        &self,
+        poll: &mut Poll,
+        additional_lifecycle_register: &mut AdditionalLifecycleEventsSet,
+        registration_token: RegistrationToken,
+    ) -> crate::Result<bool> {
         if let Ok(mut me) = self.try_borrow_mut() {
             me.source.unregister(poll)?;
+            if me.needs_additional_lifecycle_events {
+                additional_lifecycle_register.unregister(registration_token);
+            }
             Ok(true)
         } else {
             Ok(false)
         }
     }
 
-    fn pre_run(&self, data: &mut Data) -> crate::Result<()> {
+    fn before_sleep(&self) -> crate::Result<Option<(Readiness, Token)>> {
         let mut disp = self.borrow_mut();
-        let DispatcherInner {
-            ref mut source,
-            ref mut callback,
-        } = *disp;
-        source.pre_run(|event, meta| callback(event, meta, data))
+        let DispatcherInner { ref mut source, .. } = *disp;
+        source.before_sleep()
     }
 
-    fn post_run(&self, data: &mut Data) -> crate::Result<()> {
+    fn before_handle_events(&self, events: EventIterator<'_>) {
         let mut disp = self.borrow_mut();
-        let DispatcherInner {
-            ref mut source,
-            ref mut callback,
-        } = *disp;
-        source.post_run(|event, meta| callback(event, meta, data))
+        let DispatcherInner { ref mut source, .. } = *disp;
+        source.before_handle_events(events);
     }
 }
 
@@ -356,15 +397,46 @@ pub(crate) trait EventDispatcher<Data> {
         data: &mut Data,
     ) -> crate::Result<PostAction>;
 
-    fn register(&self, poll: &mut Poll, token_factory: &mut TokenFactory) -> crate::Result<()>;
+    fn register(
+        &self,
+        poll: &mut Poll,
+        additional_lifecycle_register: &mut AdditionalLifecycleEventsSet,
+        token_factory: &mut TokenFactory,
+    ) -> crate::Result<()>;
 
-    fn reregister(&self, poll: &mut Poll, token_factory: &mut TokenFactory) -> crate::Result<bool>;
+    fn reregister(
+        &self,
+        poll: &mut Poll,
+        additional_lifecycle_register: &mut AdditionalLifecycleEventsSet,
+        token_factory: &mut TokenFactory,
+    ) -> crate::Result<bool>;
 
-    fn unregister(&self, poll: &mut Poll) -> crate::Result<bool>;
+    fn unregister(
+        &self,
+        poll: &mut Poll,
+        additional_lifecycle_register: &mut AdditionalLifecycleEventsSet,
+        registration_token: RegistrationToken,
+    ) -> crate::Result<bool>;
 
-    fn pre_run(&self, data: &mut Data) -> crate::Result<()>;
+    fn before_sleep(&self) -> crate::Result<Option<(Readiness, Token)>>;
+    fn before_handle_events(&self, events: EventIterator<'_>);
+}
 
-    fn post_run(&self, data: &mut Data) -> crate::Result<()>;
+#[derive(Default)]
+/// The list of events
+pub(crate) struct AdditionalLifecycleEventsSet {
+    /// The list of sources
+    pub(crate) values: Vec<RegistrationToken>,
+}
+
+impl AdditionalLifecycleEventsSet {
+    fn register(&mut self, token: RegistrationToken) {
+        self.values.push(token)
+    }
+
+    fn unregister(&mut self, token: RegistrationToken) {
+        self.values.retain(|it| it != &token)
+    }
 }
 
 // An internal trait to erase the `F` type parameter of `DispatcherInner`
@@ -412,7 +484,7 @@ where
 pub struct Dispatcher<'a, S, Data>(Rc<dyn ErasedDispatcher<'a, S, Data> + 'a>);
 
 impl<'a, S, Data> std::fmt::Debug for Dispatcher<'a, S, Data> {
-    #[cfg_attr(coverage, no_coverage)]
+    #[cfg_attr(feature = "nightly_coverage", coverage(off))]
     fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
         f.write_str("Dispatcher { ... }")
     }
@@ -429,7 +501,11 @@ where
     where
         F: FnMut(S::Event, &mut S::Metadata, &mut Data) -> S::Ret + 'a,
     {
-        Dispatcher(Rc::new(RefCell::new(DispatcherInner { source, callback })))
+        Dispatcher(Rc::new(RefCell::new(DispatcherInner {
+            source,
+            callback,
+            needs_additional_lifecycle_events: S::NEEDS_EXTRA_LIFECYCLE_EVENTS,
+        })))
     }
 
     /// Returns an immutable reference to the event source.
@@ -485,7 +561,7 @@ pub struct Idle<'i> {
 }
 
 impl<'i> std::fmt::Debug for Idle<'i> {
-    #[cfg_attr(coverage, no_coverage)]
+    #[cfg_attr(feature = "nightly_coverage", coverage(off))]
     fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
         f.write_str("Idle { ... }")
     }
